



python -c "import PySpice; print(PySpice.__version__)"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest unit-test
IF %ERRORLEVEL% NEQ 0 exit /B 1
pyspice-post-installation --check-install
IF %ERRORLEVEL% NEQ 0 exit /B 1
setx PySpiceLibraryPath %SRC_DIR%\examples\libraries
IF %ERRORLEVEL% NEQ 0 exit /B 1
setx PYTHONLEGACYWINDOWSSTDIO utf_8
IF %ERRORLEVEL% NEQ 0 exit /B 1
setx PYTHONIOENCODING utf_8
IF %ERRORLEVEL% NEQ 0 exit /B 1
chcp 65001
IF %ERRORLEVEL% NEQ 0 exit /B 1
invoke test.run-examples
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
